//***********************************************************************************************************************************************************
//Instructions
//This code runs numerical experiments for the Gopher Fast Fourier Transform Algorithm (GFFT) algorithm. 

//Parameter setting starts at line 28 (after the include and using namespace commands).

//Read the instruction of parameter set up section and set the parameters as you want
//One can include the directory location of fftw3.h in the include command. 
//For example, one can change #include <fftw3.h> to #include </User/username/usr/include/fftw3.h> to let the compiler know the correct path of fftw3.h.

//We have to compile this file (StructuredSFT.cpp) with primes.h using Makefile. 
//***********************************************************************************************************************************************************
#include <complex>
#include <iostream>
#include <vector>
#include <map>
#include <algorithm>
#include <cmath>
#include "primes.h"
#include <fftw3.h>
#include <ctime>
#include <cstdlib>
#include <ctime>

using namespace std;


//***********************************************************************************************************************************************************
//Below is the part of code for parameter setting
//***********************************************************************************************************************************************************

//N is the bandwidth, later will take the value of N from outside (N is the smallest power of two number that greater than sample size)
//e.g. if sample size is 1000, then N is 2^10 = 1024
const int N = pow(2, 22);
//number of blocks in Fourier space
const int num_block = 3;
//length of block in Fourier space
const int length_block = pow(2, 4);
//set this to be true of want use FFTW_MEASURE, otherwise, set it to be false
bool fftwMeasure = false;
//number of repeated numerical test
int num_test = 1;
//number of different numerical test
int rounds = 10;
//set this to be true if want to run the numerical experiment in the noiseless case
//otherwise set it to be false
bool noiseless = false;
//Signal to Noise Ratio
double SNR = 30;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%







//***********************************************************************************************************************************************************
//WARNING!!!
//Do NOT edit the code below unless you know what you are doing
//***********************************************************************************************************************************************************

const double pi = acos(-1);
const complex<double> I(0, 1);

//Number of significant frequencies, later will also take this number from outside
const int k = num_block*length_block;



//calculate a^b
int power(int a, int b) {
	int temp1 = 1;
	for (int i = 0; i < b; ++i) {
		temp1 = temp1*a;
	}
	return temp1;
}


//calculate pow(a, 2)
double pow_double(double a){
	return a*a;
}


//In this function, we try to find the index of the smallest main prime number
int index_of_smallest_main_prime(int N, int k) {
	int prime_num = log(N) / log(2) + 1;  //we will need at most log(N) / log(2) + 1 prime numbers.
	vector<int> p(prime_num);
	//this vector record which prime number shows up previously
	//0 indicate does not show up; otherwise indicate shows up
	vector<int> show_up(prime_num);
	//alpha record the power of each prime numbers
	vector<int> alpha(prime_num);
	//We will do not want q shows up in our prime sequence again, since we already have it in our prime sequence
	for (int i = 0; i < prime_num; ++i) {
		p[i] = PRIMES[i];
		show_up[i] = 0;
		alpha[i] = 0;
	}
	
	int temp = 1;
	int currentmin;
	int possiblemin;
	int l;
	int l_min;
	
	int S = 1;
	//2 must shows up in the first round
	show_up[0] = 1;
	
	while (temp < N / p[S]) {
		temp = 1;
		currentmin = N;
		l = 0;
		l_min = 0;
		while (alpha[l] > 0) {
			possiblemin = power(p[l], alpha[l])*(p[l] - 1);
			if (possiblemin < currentmin) {
				currentmin = possiblemin;
				l_min = l;
			}
			++l;
		}
		if (p[l] < currentmin) {
			l_min = l;
			alpha[l_min] = 1;
			++show_up[l_min];
		}
		else {
			++alpha[l_min];
			++show_up[l_min];
		}
		for (int i = 0; i < prime_num; ++i) {
			temp = temp*power(p[i], alpha[i]);
		}
		//update S the be the smallest index in show_up that is equal to 0
		for (int i = 0; i < show_up.size(); ++i){
			if (show_up[i] == 0){
				S = i;
				break;
			}
		}
	}
	if (PRIMES[S] >= k){
		return S;
	}
	else{
		while (PRIMES[S] < k){
			++S;
		}
		return S;
	}
	
}



//In this part we try to build a sequence of prime numbers that we will use in our sampling scheme.
//Here we follow the principle that we want product of the sequence p_l^(alpha_l) goes to N/S as slow as possible.
//More specifically, the scheme we have will be q_j^(alpha_j), q_j^(alpha_j)*p_j1^(beta_j1), q_j^(alpha_j)*p_j1^(beta_j1)*p_j2^(beta_j2)....
//This prime sequence should satisfy q_j^(alpha_j)*p_j1^(beta_j1)*p_j2^(beta_j2)...>=N
//Also we want sum(q_j^(alpha_j), p_j1^(beta_j1), p_j2^(beta_j2)...) to ba as small as possible
//And for each p_j1^(beta_j1), p_j2^(beta_j2)... they should be close to each other as much as possible.
//The last property will be very useful later when we make our algorithm parallel.
void PrimeSequence(vector<int>& prime, vector<int>& power_of_prime, int q, int N) {
	
	prime.push_back(q);
	power_of_prime.push_back(1);
	
	int S = power(q, power_of_prime[0]);
	int prime_num = log(N / S) / log(2) + 1;  //we will need at most log(N / S) / log(2) + 1 prime numbers.
	vector<int> p(prime_num);


	//We will do not want q shows up in our prime sequence again, since we already have it in our prime sequence
	int q_occurs = 0;
	for (int i = 0; i < prime_num; ++i) {
		if (PRIMES[i] != q) {
			p[i] = PRIMES[i + q_occurs];
		}
		else {
			++q_occurs;
			p[i] = PRIMES[i + q_occurs];
		}
	}
	vector<int> alpha(prime_num);
	for (int i = 0; i < prime_num; ++i) {
		alpha[i] = 0;
	}
	int temp = 1;
	int currentmin;
	int possiblemin;
	int l;
	int l_min;
	while (temp < N / S) {
		temp = 1;
		currentmin = N;
		l = 0;
		l_min = 0;
		while (alpha[l] > 0) {
			possiblemin = power(p[l], alpha[l])*(p[l] - 1);
			if (possiblemin < currentmin) {
				currentmin = possiblemin;
				l_min = l;
			}
			++l;
		}
		if (p[l] < currentmin) {
			l_min = l;
			alpha[l_min] = 1;
		}
		else {
			++alpha[l_min];
		}
		for (int i = 0; i < prime_num; ++i) {
			temp = temp*power(p[i], alpha[i]);
		}
	}
	//for (int i = 0; i < prime_num; ++i) {
	//std::cout << alpha[i] << std::endl;
	//}
	for (int i = 0; i < prime_num; ++i) {
		if (alpha[i] != 0) {
			prime.push_back(p[i]);
			power_of_prime.push_back(alpha[i]);
		}
	}
	//return alpha;
}


//In this function we calculate the value for phi function in Fermat-Euler Theorem by using Euler's product formula first, 
//Then return the value of phi function - 1, we will then use this value to calculate the modular inverse.
//Here the input n = q^(alpha)*p_1^(beta_1)*p_2^(beta_2)...*p_j^(beta_j)
//prime is a vector contains q, p_1, p_2, p_j
int Phi_n_1(int n, vector<int> prime) {
	int length = prime.size();
	int answer = n;
	for (int i = 0; i < length; ++i) {
		answer = answer / prime[i] * (prime[i] - 1);
	}
	return answer - 1;
}





//In this part we use FermatEuler theorem to find the inverse (modulo ) for 'a mod n'. 
//The inverse (modulo) for 'a mod n' is defined in this way: if 'b' is inverse (modulo) for 'a mod n', 
//then b*a = 1 mod n. 
//The value of phi_n_1 comes from Phi_n_1 function, here the phi function is the one in Fermat-Euler Theorem
//In this part, we use long long int to prevent overflow. 
long long int InverseModP(int a, int phi_n_1, int n)
{
	long long int y = a;
	long long int temp = 1;
	
	//Here we apply the property of Euler's totient function. 
	//When q is a prime nbumber, the Euler's totient function of q^alpha is equal to q^(alpha)-q^(alpha-1)
	long long int temp1 = phi_n_1;
	//Since it will overflow if we calculate a^(temp1), so we will express temp1 into base 2 and then use the
	//equation: ab mod n = (a mod n)(b mod n) mod n
	while (temp1 != 0)
	{
		if (temp1 % 2 == 1)   //check whether it's current digit in base 2 is 1
		{
			temp = (temp*y) % n;
		}
		y = (y*y) % n;
		temp1 = temp1 / 2;
	}
	return temp%n;
}


//In this part we create one row for sample scheme matrix, it is in the form:
//q_i^(alpha_i)  p_i1^(beta_i1)  p_i2^(beta_i2), ..., p_ik^(beta_ik)
void row_sampling_scheme(vector<vector <int> >& sample_scheme, vector<int> prime, vector<int>alpha) {
	vector<int> temp;   //use temp to store value for each row and then push_back temp in to sample_scheme
	int length_prime = prime.size();
	int temp1 = power(prime[0], alpha[0]);
	temp.push_back(temp1);
	for (int i = 1; i < length_prime; ++i) {
		temp.push_back(power(prime[i], alpha[i]));
	}
	//Add a row in sample_scheme.
	sample_scheme.push_back(temp);
	// cout << temp.size() << endl;
}


//create struct, struct contains four elements: 
//one is p, it is equal to the value of p_ij^(beta_ij)
//one is qp, it is equal to the vaue of  q_i^(alpha_i)*p_i1^(beta_i1)*...*p_i(j-1)^(beta_i(j-1))
//one is pInverseqp, it is equal to p_ij^(beta_ij) inverse mod q_i^(alpha_i)*p_i1^(beta_i1)*...*p_i(j-1)^(beta_i(j-1))
struct recovery_info
{
	int p;
	int qp;
	long long int pInverseqp;
};



//Generate sample_scheme, which is in the form:
//q_1^(alpha_1)  p_11^(beta_11)  p_12^(beta_12), ..., p_1k^(beta_1k)
//.
//.
//.
//q_m^(alpha_m)  p_m1^(beta_m1)  p_m2^(beta_m2), ..., p_mk^(beta_mk')
//The matrix contains m rows, but for each row, the number of elements may be different.
//Here m is a number relative to sparsity, N is the bandwidth
void sampling_scheme_recovery_info(vector<vector <int> >& sample_scheme, vector<vector <recovery_info> >& Recovery_Info, int m, int N, int k) {
	int phi;
	fftw_complex *in, *out;
	
	int smallest_main_index = index_of_smallest_main_prime(N, k);
	
	for (int i = 0; i < m; ++i) {
		//we will use temp1 to store subvector of prime, for each loop, we initialize temp1 first in order to set it as an empty vector.
		vector<int> temp;
		//generate prime sequence and their power first
		vector<int> prime;
		vector<int> power_of_prime;
		//Here we use PrimeSequence function generate prime and power_of_prime, they are vectors with same length
		PrimeSequence(prime, power_of_prime, PRIMES[smallest_main_index + i], N);
		//use row_sampling_scheme generate a row of sample scheme matrix and push it back into sample_scheme
		row_sampling_scheme(sample_scheme, prime, power_of_prime);

		Recovery_Info.push_back(vector<recovery_info>());

		//Build Recovery_Info
		//temp1 is the pointer to the i th row of Recovery_Info
		vector<recovery_info> &temp1 = Recovery_Info[i];
		temp1.push_back(recovery_info());
		//p_i1^(beta_i1)
		temp1[0].p = sample_scheme[i][1];
		//q_i^(alpha_i)
		temp1[0].qp = sample_scheme[i][0];
		temp.push_back(prime[0]);
		//here phi is the short version for phi(n) - 1
		phi = Phi_n_1(sample_scheme[i][0], temp);
		//p_i1^(beta_i1) inverse mod q_i^(alpha_i)
		temp1[0].pInverseqp = InverseModP(sample_scheme[i][1], phi, sample_scheme[i][0]);
		
		//So far, we have first struct in row i
		//In this for loop, we will push_back other structs in row i of Recoery_Info
		for (int j = 1; j < prime.size() - 1; ++j) {
			temp1.push_back(recovery_info());
			temp1[j].p = sample_scheme[i][j+1];
			temp1[j].qp = temp1[j - 1].qp*temp1[j - 1].p;
			temp.push_back(prime[j]);
			phi = Phi_n_1(temp1[j].qp, temp);
			temp1[j].pInverseqp = InverseModP(temp1[j].p, phi, temp1[j].qp);
		}
	}
}



//generate fftw_plans as a dictionary: map <int, fftw_plan> fftw_plans
//the element of fftw_plans is a pair, for which contains two part:
//int: corresponding to the length of DFT
//fftw_plan: store the pointer of fftw_plan with corresponding length
void create_plans(map <int, fftw_plan>& fftw_plans, map <int, fftw_plan>& fftw_plans_back, vector<int>& length){
	
	if(fftwMeasure){
		fftw_complex *in, *out;
	    for (int i = 0; i < length.size(); ++i){
		    in = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    fftw_plans.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_FORWARD, FFTW_MEASURE)));
            fftw_plans_back.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_BACKWARD, FFTW_MEASURE)));
	    }
	}
	
	else{
	    cout << "Using FFTW_ESTIMATE, not FFTW_MEASURE!" << endl;
	    fftw_complex *in, *out;
	    for (int i = 0; i < length.size(); ++i){
		    in = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    fftw_plans.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_FORWARD, FFTW_ESTIMATE)));
            fftw_plans_back.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_BACKWARD, FFTW_ESTIMATE)));
	    }
	}
}




void create_A_hat_stu(map <int, fftw_complex*>& DFT_samples, vector<int> length, vector<double>& frequency, vector<complex<double>>& coefficient){
	fftw_complex* temp;
	for(int i = 0; i < length.size(); ++i){
		temp = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		for (int k = 0; k < length[i]; ++k){
			temp[k][0] = 0;
			temp[k][1] = 0;
		}
		//fill temp
		for (int j = 0; j < frequency.size(); ++j){
			if (frequency[j] >= 0){
				temp[int(frequency[j]) % length[i]][0] += real(coefficient[j]);
			    temp[int(frequency[j]) % length[i]][1] += imag(coefficient[j]);
			}
			else{
				temp[(int(frequency[j]) % length[i]) + length[i]][0] += real(coefficient[j]);
			    temp[(int(frequency[j]) % length[i]) + length[i]][1] += imag(coefficient[j]);
			}
			//temp[int(frequency[j]) % length[i]][0] += real(coefficient[j]);
			//temp[int(frequency[j]) % length[i]][1] += imag(coefficient[j]);
		}
		DFT_samples.insert(make_pair(length[i], temp));
	}
}



//A_hat ---> samples
void create_sample_time(vector<fftw_complex*>& samples, vector<int>& length, map <int, fftw_complex*>& A_hat, map <int, fftw_plan>& fftw_plans_back){
	complex<double> noise;
	double u1, v1;
	
	for (int i = 0; i < length.size(); ++i){
		fftw_execute_dft(fftw_plans_back[length[i]], A_hat[length[i]], samples[i]);
		/*
		//add noise
		for (int j = 0; j < length[i]; ++j){
			u1 = drand48();
            v1 = drand48();
	        //create random noise
	        noise = std_noise * sqrt(-2*log(u1)) * exp(2*M_PI * I * v1);
			
			samples[i][j][0] += real(noise);
			samples[i][j][1] += imag(noise);
		}
		*/
	}
}


void add_noise(vector<fftw_complex*>& samples, vector<int>& length){
	//assume the total number of elements in samples is m
	int total_length = 0;
	double sig_power = 0;
	double noise_power = 0;
	//evaluate the power of all the elements in samples
	for(int i = 0; i < samples.size(); ++i){
		total_length  += length[i];
		for(int j = 0; j < length[i]; ++j){
			sig_power += (samples[i][j][0]*samples[i][j][0] + samples[i][j][1]*samples[i][j][1]);
		}
	}
	
	//create a noise vector which contains m elements
	vector<complex<double>> noise_vector;
	noise_vector.resize(total_length);
	//each elements is a i.i.d random Gaussian noise with mean 0 and variance 1
	//using box muller method
	double u1, v1;
	for(int i = 0; i < total_length; ++i){
		u1 = drand48();
        v1 = drand48();
		noise_vector[i] = sqrt(-2*log(u1)) * exp(2*M_PI * I * v1);
		//calculate the power of noise vector
		noise_power += -2*log(u1);
	}
	
	//normalize the noise vector based on desired SNR value, the power of noise vector
	//and the power of samples
	int temp = 0;
	double norm_factor = sqrt(sig_power/double(pow(10, SNR/10)))/sqrt(noise_power);
	for(int i = 0; i < total_length; ++i){
		noise_vector[i] *= norm_factor;
	}
	
	
	//exam SNR
	double new_noise_power = 0;
	for(int i = 0; i < total_length; ++i){
		new_noise_power += (abs(noise_vector[i])*abs(noise_vector[i]));
	}
	//cout << "noise_power is: " << noise_power << endl;
	//cout << "norm_factor is: " << norm_factor << endl;
	//cout << "sig_power is: " << sig_power << endl;
	//cout << "new_noise_power is: " << new_noise_power << endl;
	cout << "The actual SNR is: " << 10*log10(sig_power/new_noise_power) << endl;;
	
	
	//add noise vector with samples elementwise
	int noise_vector_index = 0;
	for(int i = 0; i < samples.size(); ++i){
		for(int j = 0; j < length[i]; ++j){
			samples[i][j][0] += real(noise_vector[noise_vector_index]);
			samples[i][j][1] += imag(noise_vector[noise_vector_index]);
			++noise_vector_index;
		}
	}
	
}

//calculate Discrete Fourier Transform by using FFTW
//input: vector<fftw_complex*>& samples (vector that cotains all the samples)
//input: map <int, fftw_plan>& fftw_plans, this is the 
//inout: vector<int>& sample_scheme_flat: a vector stores the length of all needed sequence in MSFT
//output: map <int, fftw_complex*>& DFT_samples (vector that contains all the samples after Fourier TRansform)
void DFT(vector<fftw_complex*>& samples, map <int, fftw_plan>& fftw_plans, map <int, fftw_complex*>& DFT_samples, vector<int>& sample_scheme_flat) {
	fftw_complex *in, *out;
	//for each DFT define
	//1. define in and out
	//2. record the length of each vector that will take DFT
	//3. execute corresponding fftw_plan to get out
	//4. store out in DFT_samples
	int dft_length;
	for (int i = 0; i < samples.size(); ++i) {
		dft_length = sample_scheme_flat[i];
		//1. define in and out
		in = samples[i];
		out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * dft_length);
		
		
		fftw_execute_dft(fftw_plans[dft_length], in, out);
		
		for (int l = 0; l < dft_length; ++l){
			out[l][0] = out[l][0]/dft_length;
			out[l][1] = out[l][1]/dft_length;
		}
		
		DFT_samples.insert(make_pair(dft_length, out));
	}
}


//define own function to compare two pairs <double number, index(integer)>
bool mag_complex(pair <complex<double>, int> a, pair <complex<double>, int> b)
{
	return abs(a.first) > abs(b.first);
}

//convert data structure from fftw_complex to std.complex
complex<double> fftw_to_std(fftw_complex c){
	return {c[0], c[1]};
}

//find index of t_min in algorithm 2
//input: A_sort[b], DFT_samples[p[l]*p[q+j]], p[l], p[q+j], r[0]
int t_m(complex <double> sorted_element, fftw_complex* DFT_vec, double p_l, double p_qj, int r_0) {
	double temp_min = abs(sorted_element - fftw_to_std(DFT_vec[r_0]));
	int t_m = 0;
	for (int t = 1; t < p_l; ++t) {
		if (abs(sorted_element - fftw_to_std(DFT_vec[t*(int)p_qj + r_0])) < temp_min) {
			temp_min = abs(sorted_element - fftw_to_std(DFT_vec[t*(int)p_qj + r_0]));
			t_m = t;
		}
	}
	return t_m;
}



//find median of vector
double findMedian(vector<double> vec) {
	double median;
	int size_vec = vec.size();

	sort(vec.begin(), vec.end());

	if (size_vec % 2 == 0) {
		median = (vec[size_vec / 2 - 1] + vec[size_vec / 2]) / 2.0;
	}
	else {
		median = vec[size_vec / 2];
	}
	return median;
}


//Recover frequency using information in Recovery_Info
//try torecover frequency x satisfy: rem[i] = freq % num[i]
//line_num tells which line of information of Recovery_Info should be used here
int findMinFreq(vector<int>& rem, vector<vector <recovery_info> >& Recovery_Info, int line_num){
	int x = rem[0];
	for (int i = 1; i < rem.size(); ++i){
		x = (rem[i] + (x - rem[i]) * Recovery_Info[line_num][i-1].pInverseqp * Recovery_Info[line_num][i-1].p) % (Recovery_Info[line_num][i-1].qp * Recovery_Info[line_num][i-1].p);
		if (x < 0){
			x = (Recovery_Info[line_num][i-1].qp * Recovery_Info[line_num][i-1].p) + x;
		}
	}
	return x;
}


//Energetic Frequencies Identification
//Find energetic frequencies in effective support of current window function
//input: 
//vector<vector <recovery_info> >& Recovery_Info: stores the inverse modular infomation that cna be used to recover significant frequencies according to Chinese Reminder Theorem
//map <int, int>& C: stores the information of all recovered frequencies. The first numnber of the frequency; the second number is how many times it has been recovered.
//vector<vector <int> >& sample_scheme: smapling scheme of MSFT. 
//map <int, fftw_complex*>& DFT_samples: DSFT transform of needed sequences
//int filter_shift: current center of periodic Gaussian function
//int FilterShift: floor(N / double(filter_num))
void EFI(vector<vector <recovery_info> >& Recovery_Info, map <int, int>& C, vector<vector <int> >& sample_scheme, map <int, fftw_complex*>& DFT_samples, int B){
	//the element of A_sort is a pair. In the pair the first element is a complex number, the second element is the index based on sort result
	vector <pair <complex<double>, int>> A_sort;
	vector <int> r;
	int t_min;
	int temp_C;
	
	for (int j = 0; j < sample_scheme.size(); ++j) {
		//initialize a new vector A_sort to store the sorted complex vector
		A_sort.clear();
		for (int i = 0; i < sample_scheme[j][0]; ++i) {
			const auto &c = DFT_samples[sample_scheme[j][0]][i];
			A_sort.push_back(make_pair(complex<double>{c[0], c[1]}, i));
		}
		sort(A_sort.begin(), A_sort.end(), mag_complex);
		for (int b = 0; b < B; ++b) {
			r.clear();
			//no need to add index by 1, since index start from zero
			r.push_back(A_sort[b].second);
			for (int l = 1; l < sample_scheme[j].size(); ++l) {
				t_min = t_m(A_sort[b].first, DFT_samples[sample_scheme[j][0] * sample_scheme[j][l]], sample_scheme[j][l], sample_scheme[j][0], r[0]);
				r.push_back(int(r[0] + t_min*sample_scheme[j][0]) % int(sample_scheme[j][l]));
			}
			
			//Use Chinese Reminder Theorem to reconstruct the frequency
			temp_C = findMinFreq(r, Recovery_Info, j) % N;
			
			//count how many times this frequency has been found
			if (C.find(temp_C) == C.end()) {
				//not found yet
				C.insert(pair<int, int>(temp_C, 1));
			}
			else {
				//found before
				C[temp_C] += 1;
			}
		}
	}
}


//Coefficient Estimation
//CE(R, C, sample_scheme, DFT_samples, c, filter_shift)
//R is the place store the final result, i.e. the frequnecy and coefficient pairs
//C contains the <frequency, show up times> pair
//sample_scheme contains the information of sampling scheme, here we use it to find the correct position of DFT_smaples
//DFT_samples contains the sequences in Fourier space
//c is the parameter in periodic Gaussian function
//filter_shift contains the information of corrent center of periodic Gaussian funciton, we use this to scale the coefficient from G*f to f
void CE(vector <pair<int, complex<double>>>& R, map <int, int>& C, vector<vector <int> >& sample_scheme, map <int, fftw_complex*>& DFT_samples){
	
	//create a vector to store the selected elements from A[p[m-1]*p[q+h]]
	vector<double> real_part, imag_part;
	//create a vector with pair as elements to store the significant frequencies
	
	complex<double> temp_coeff;
	
	
	for (const auto &item : C) {
		
		if (item.second >= 2) {
			real_part.clear();
			imag_part.clear();
			
			for (int h = 0; h < sample_scheme.size(); ++h) {
				real_part.push_back(DFT_samples[sample_scheme[h][0] * sample_scheme[h].end()[-1]][item.first % int(sample_scheme[h][0] * sample_scheme[h].end()[-1])][0]);
				imag_part.push_back(DFT_samples[sample_scheme[h][0] * sample_scheme[h].end()[-1]][item.first % int(sample_scheme[h][0] * sample_scheme[h].end()[-1])][1]);
				real_part.push_back(DFT_samples[sample_scheme[h][0] * sample_scheme[h].end()[-2]][item.first % int(sample_scheme[h][0] * sample_scheme[h].end()[-2])][0]);
				imag_part.push_back(DFT_samples[sample_scheme[h][0] * sample_scheme[h].end()[-2]][item.first % int(sample_scheme[h][0] * sample_scheme[h].end()[-2])][1]);
			}
			
			temp_coeff = findMedian(real_part) + findMedian(imag_part)*I;
			
			//set the coefficient to be the median of real_part and imag_part
			R.push_back(pair<int, complex<double>>(item.first, temp_coeff));
		}
	}
}



//define own function to compare two pairs in R
bool R_mag_complex(pair<int, complex<double>> a, pair<int, complex<double>> b)
{
	return abs(a.second) > abs(b.second);
}

//define own function to compare two pairs in R
bool R_freq(pair<int, complex<double>> a, pair<int, complex<double>> b)
{
	return abs(a.first) > abs(b.first);
}


int main()
{
	//in rounds number of numerical experiments, err_num of them fails to find all the significant frequencies
	int err_num = 0;
	//later will be use to store the largest error (abs(true - extimate)/abs(true)) in current round of numerical experiments
	double err = 0;
	
	double avg_l1_err = 0;
	
	const int B = k;
	int temp = 1;
	int m_start = index_of_smallest_main_prime(N, k);
	int m = 0;
		
	//m is the number of set of prime numbers that we will need in sampling_scheme
	while (temp < N) {
		temp = temp*PRIMES[m_start + m];
		++m;
	}
	//set the smallest number of set of samples to increase stability
	
	//This is the randomized setting
	//m = (m < ceil(number_of_sample_sets*log(k))) ? ceil(number_of_sample_sets*log(k)) : m;
	
	//This is the deterministics setting
	if (floor(log(N)/log(k)) > 1){
		m = 2*k*floor(log(N)/log(k)) + 1;
	}
	else{
		m = 2*k + 1;
	}
	
	//Sample scheme
	vector<vector <int> > sample_scheme;
	//Recovery information
	vector<vector <recovery_info> > Recovery_Info;
	//q_1^(alpha_1)  q_1^(alpha_1)*p_11^(beta_11)  q_1^(alpha_1)*p_12^(beta_12), ..., q_1^(alpha_1)*p_1k^(beta_1k)
	//is one set of sample scheme
	//cout << "Start create sampling scheme and recovery information..." << '\n';
	sampling_scheme_recovery_info(sample_scheme, Recovery_Info, m, N, k);
	
	int total_samples = 0;
	//cout << "Below is the sampling scheme matrix:" << '\n';
	for (int i = 0;i < sample_scheme.size(); ++i){
		total_samples = total_samples + sample_scheme[i][0];
		for (int j = 0; j < sample_scheme[i].size(); ++j){
			//cout << sample_scheme[i][j] << " ";
			if (j > 0){
				total_samples = total_samples + sample_scheme[i][0]*sample_scheme[i][j];
			}
		}
		//cout << '\n';
	}
	cout << "Sample to bandwidth ratio is: " << total_samples / double(N) << endl;
	
	//flatten the sample scheme. Later will be used as the set of key to find the correct position in some data structures like DFT_smaples
	vector<int> sample_scheme_flat;
	for (int i = 0;i < sample_scheme.size(); ++i){
		sample_scheme_flat.push_back(sample_scheme[i][0]);
		for (int j = 1; j < sample_scheme[i].size(); ++j){
			sample_scheme_flat.push_back(sample_scheme[i][0]*sample_scheme[i][j]);
		}
	}
	
	cout << "create fftw plans for forward and backward FFT" << endl;
	//Compute the fftw_plan for different length of samples
	map <int, fftw_plan> fftw_plans;
	map <int, fftw_plan> fftw_plans_backward;
	//int stores the length of each fftw_plan, since all numbers are pairwise prime, each length must be unique
	create_plans(fftw_plans, fftw_plans_backward, sample_scheme_flat);
	//create_plans_backward(fftw_plans_backward, sample_length);
	
	
	//samples contains all the samples that will be needed in MSFT
	//at this point we only initialize samples based on sample_scheme
	//the value of each element in samples will be filled in later
	//each element of samples is a fftw_complex pointer, fftw_complex is a high efficiency data structure of fftw3
	cout << "Initialize samples..." << endl;
	//initialize samples
	vector<fftw_complex*> samples;
	fftw_complex *samples_temp;
	for (int k = 0; k < sample_scheme_flat.size(); ++k){
		samples_temp = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * sample_scheme_flat[k]);
		samples.push_back(samples_temp);
	}
	
	vector<double> err_record;
	
	double duration;
	
	int seed = 123;
	cout << "Start numerical tests..." << endl;
	//start different numerical test
	for (int round = 0; round < rounds; ++round){
	
	srand(seed);
	++seed;
	vector<double> freq;
	vector<complex<double>> coeff (k,0);
	//create random frequencies and coefficients
	vector<double>::iterator it;
	while (freq.size() < k)
	{
		for (int i = 0; i < k - freq.size(); i++)
			freq.push_back(rand() % N);
		sort(freq.begin(), freq.end());
		it = unique(freq.begin(), freq.end());
		freq.resize(it - freq.begin());
	}
	
	
	double arg;
	for (int i = 0; i < k; ++i){
		arg = 2 * pi * drand48();
		//it is worth to mention that below we do not need to normalize the coefficient with abs(complex<double>(cos(arg), sin(arg)))
		coeff[i] = /*rad*/complex<double>(cos(arg), sin(arg))/abs(complex<double>(cos(arg), sin(arg)));
	}
	
	map <int, fftw_complex*> A_hat;
	
	//cout << "Calculate A_hat..." << endl;
	create_A_hat_stu(A_hat, sample_scheme_flat, freq, coeff);
	
	//cout << "Create needed samples in time space..." << endl;
	
	//create needed samples
	create_sample_time(samples, sample_scheme_flat, A_hat, fftw_plans_backward);
	
	if(!noiseless){
		add_noise(samples, sample_scheme_flat);
	}
	
	//R is where store the frequencies and coefficients recovered from one filter (window) function
	//pair<frequency, coefficient>
	vector <pair<int, complex<double>>> R;
	//pick k frequencies with largest magniture from R
	//Store it in R_final
	vector <pair<int, complex<double>>> R_final;
	//convert freq and coeff into the same form of R_final
	vector <pair<int, complex<double>>> R_true;
	
	map <int, fftw_complex*> DFT_samples;
	map <int, int> C;
	clock_t start;
	
	start = std::clock();
	//repeated numerical test
	for (int boo = 0; boo < num_test; ++boo){
	
	
	    R.clear();
	    R_final.clear();
	    R_true.clear();		
		
	    //How can we design a data structure that can store samples' informatoin efficiently (both samples and size of samples)
	    //sol: map (dictionary)
	    
	    //Use FFTW to get DFT samples
	    DFT(samples, fftw_plans, DFT_samples, sample_scheme_flat);
		
		//Energetic frequency identification
	    EFI(Recovery_Info, C, sample_scheme, DFT_samples, B);
		
	    //For CE (coefficient estimation) function:
	    //inputs are: R, C, sample_scheme, DFT_samples, m, c
	    CE(R, C, sample_scheme, DFT_samples);
		
		C.clear();
		DFT_samples.clear();
		
	}
	//end num_test
	
	//calculate duration of repeat numerical experiment
	duration = ( clock() - start ) / (double) CLOCKS_PER_SEC;
	
	
	//only select k frequencies with highest magnitude of coefficient
	sort(R.begin(), R.end(), R_mag_complex);
	for (int i = 0; i < k; ++i) {
		R_final.push_back(R[i]);
	}
	
	
	//convert true freq and coeff into the same form of R_final
	//varify results
	for (int i = 0; i < k; ++i){
		R_true.push_back(pair<int, complex<double>>(freq[i], coeff[i]));
	}
	
	//sort R and R_true
	sort(R_final.begin(), R_final.end(), R_freq);
	sort(R_true.begin(), R_true.end(), R_freq);
	
	//check whether it is a correct recovery
	bool TRUE_SET = true;
	//err and err_num
	for (int i = 0; i < k; ++i){
		if (R_final[i].first != R_true[i].first){
			err_num += 1;
			TRUE_SET = false;
			break;
		}
	}
	//if it is a correct recovery, calculate the error
	if (TRUE_SET){
		for (int i = 0; i < k; ++i){
			avg_l1_err = avg_l1_err + abs(R_final[i].second - R_true[i].second)/k;
			if (abs(R_final[i].second - R_true[i].second)/abs(R_true[i].second) > err){
				err = abs(R_final[i].second - R_true[i].second)/abs(R_true[i].second);
			}
		}
		err_record.push_back(err);
		err = 0;
	}
	
	freq.clear();
	coeff.clear();
	
	}
	//end rounds
	
	//calculate error in average L1 norm
	double avg_err = 0;
	for (int i = 0; i < err_record.size(); ++i){
		//cout << err_record[i] << endl;
		avg_err += err_record[i];
	}
	avg_err = avg_err/err_record.size();
	
	//output results
	cout << "Bandwidth is: " << N << '\n';
	cout << "Sparsity is: " << k << '\n';
	if (!noiseless){
	    cout << "The noise level (SNR) is: " << SNR << endl;
	}
	else{
		cout << "There is no noise in the original signal." << endl;
	}
	cout << "Running time (w/o create equispaced samples) is: " << duration/num_test << "seconds" << '\n';
	cout << "Amoung " << rounds << " of numerical experiments, " << err_num << " of them are wrong." << '\n';
	cout << "The average infinite norm error is: " << avg_err << endl;
	cout << "The average L1 norm error is: " << avg_l1_err / rounds << endl;
}













