//***********************************************************************************************************************************************************//
//Instructions
//This code runs numerical experiments for the Fourier Algorithm for Structured SparsiTy (FAST).

//Parameter setting starts at line 32 (after the include and using namespace commands).

//Set up the N, num_block and B correctly. 
//Set Randomized to be true if you want to run the randomized version, i.e., FASTR.
//When running the randomized version, only increase the value of randomize_int if you want to increase the probability of correct recovery. 
//In the ReadMe file we provide exemplary test runs that demonstrate how we set the parameters.
//Set fftwMeasure to be true if you want FFTW run all the DFTs using FFTW_MEASURE plan.
//This can make the algorithm have a better running time performance, but it can take relatively long to generate fftw_plan.
//One can also include the directory location of the fftw3.h in the include command.
//For example, one can change #include <fftw3.h> to #include </User/username/usr/include/fftw3.h> to let the compiler know the correct path of fftw3.h.

//We have to compile this file (StructuredSFT.cpp) with primes.h using Makefile. 
//***********************************************************************************************************************************************************//
#include <complex>
#include <iostream>
#include <vector>
#include <map>
#include <algorithm>
#include <cmath>
#include "primes.h"
#include <ctime>
#include <cstdlib>
#include <fftw3.h>


using namespace std;

//********************************************************************************************************************//
//Below is the part of the code for setting the parameters.
//********************************************************************************************************************//
//Number of repeated numerical tests.
int num_test = 1;
//Number of different numerical tests.
int rounds = 10;
//N is the bandwidth; later we will take the value of N from outside (N is the smallest power of two that is greater than the sample size),
//e.g., if the sample size is 1000, then N is 2^10 = 1024.
const int N = pow(2, 22);
//Number of significant frequencies in each block.
const int B = pow(2,4);
//num_block is n in the pseudo-code of the (n,B)-block Fourier Algorithm for Structured SparsiTy.
const int num_block = 3;
//Set Randomized to true if you want to use the randomized version, i.e., FASTR.
//The deterministic version is FAST.
bool Randomized = false;
//Set Randomize_int to be a larger odd number if you want a higher probability of correct recovery.
int randomize_int = 5;
//Set fftwMeasure to true if you want to use FFTW_MEASURE plan (better performance, but it takes a really long time to generate the fftw_plans).
//Set it to false if you want to use FFTW_ESTIMATE plan (sub-optimal performance, but it takes much less time to generate the fftw_plans).
bool fftwMeasure = false;
//Set noiseless to true if you want to run the algorithm in the noiseless case, i.e., FAST.
//Otherwise set it to true.
bool noiseless = false;
//Signal to Noise Ratio (SNR).
double SNR = 30;


//***********************************************************************************************************************************************************//
//WARNING!!!
//Do NOT edit the code below unless you know what you are doing!
//***********************************************************************************************************************************************************//

const double pi = acos(-1);
const complex<double> I(0, 1);

//The value of u depends on B.
int u;
//Determines how many prime numbers we need in s_k.
int K;
//Corresponds to the s in Algorithm 2.
vector<int> s_k;


//Calculate a^b.
int power(int a, int b) {
	int temp1 = 1;
	for (int i = 0; i < b; ++i) {
		temp1 = temp1*a;
	}
	return temp1;
}



//In this part we try to build a sequence of prime numbers that we will use in our sampling scheme.
//Here we follow the principle that we want the product of the sequence p_l^(alpha_l) to go to N/S as slowly as possible.
//More specifically, our scheme will be q_j^(alpha_j), q_j^(alpha_j)*p_j1^(beta_j1), q_j^(alpha_j)*p_j1^(beta_j1)*p_j2^(beta_j2)... .
//This prime sequence should satisfy q_j^(alpha_j)*p_j1^(beta_j1)*p_j2^(beta_j2)...>=N.
//Also we want that sum(q_j^(alpha_j), p_j1^(beta_j1), p_j2^(beta_j2)...) is as small as possible.
//And for each p_j1^(beta_j1), p_j2^(beta_j2)... they should be close to each other as much as possible.
//The last property will be very useful later when we parallelize our algorithm.
void PrimeSequence(vector<int>& prime, vector<int>& power_of_prime, int q) {
	
	double S = u*num_block;
	int prime_num = log(N / S) / log(2) + 1;  //We will need at most log(N / S) / log(2) + 1 prime numbers.
	vector<int> p(prime_num);


	//We do not want q to show up in our prime sequence again, since we already have it in our prime sequence.
	int q_occurs = 0;
	for (int i = 0; i < prime_num; ++i) {
		if (PRIMES[i] != q) {
			p[i] = PRIMES[i + q_occurs];
		}
		else {
			++q_occurs;
			p[i] = PRIMES[i + q_occurs];
		}
	}
	vector<int> alpha(prime_num);
	for (int i = 0; i < prime_num; ++i) {
		alpha[i] = 0;
	}
	int temp = 1;
	int currentmin;
	int possiblemin;
	int l;
	int l_min;
	while (temp < N / S) {
		temp = 1;
		currentmin = N;
		l = 0;
		l_min = 0;
		while (alpha[l] > 0) {
			possiblemin = power(p[l], alpha[l])*(p[l] - 1);
			if (possiblemin < currentmin) {
				currentmin = possiblemin;
				l_min = l;
			}
			++l;
		}
		if (p[l] < currentmin) {
			l_min = l;
			alpha[l_min] = 1;
		}
		else {
			++alpha[l_min];
		}
		for (int i = 0; i < prime_num; ++i) {
			temp = temp*power(p[i], alpha[i]);
		}
	}
	for (int i = 0; i < prime_num; ++i) {
		if (alpha[i] != 0) {
			prime.push_back(p[i]);
			power_of_prime.push_back(alpha[i]);
		}
	}
	//return alpha;
}




//In this part we create a row for the sampling scheme matrix. It is of the form:
//q_i^(alpha_i)  p_i1^(beta_i1)  p_i2^(beta_i2), ..., p_ik^(beta_ik).
void row_sampling_scheme(vector<vector <int> >& sample_scheme, vector<int> prime, vector<int>alpha) {
	vector<int> temp;   //Use temp to store the value for each row and then push_back temp into sample_scheme.
	int length_prime = prime.size();
	int temp1 = power(prime[0], alpha[0]);
	temp.push_back(temp1);
	for (int i = 1; i < length_prime; ++i) {
		temp.push_back(power(prime[i], alpha[i]));
	}
	//Add a row to sample_scheme.
	sample_scheme.push_back(temp);
	// cout << temp.size() << endl;
}


//Find the smallest power of prime_num that is greater than B.
int smallest_u(int prime_num){
	int temp = prime_num;
	while (temp < B){
		temp = temp * prime_num;
	}
	return temp;
}


//Generate sample_scheme, which is of the form:
//q_1^(alpha_1)  p_11^(beta_11)  p_12^(beta_12), ..., p_1k^(beta_1k)
//.
//.
//.
//q_m^(alpha_m)  p_m1^(beta_m1)  p_m2^(beta_m2), ..., p_mk^(beta_mk').
//The matrix contains m rows, but for each row the number of elements may be different.
//In this algorithm for block structured functions we only need one set of q and p.
void t_l_info(vector<vector <int> >& sample_scheme) {
	int phi;
		
	//Here the upper index of this for loop is 1, it might need to be changed for the general, polynomially structured sparse case.
	for (int i = 0; i < 1; ++i) {
		//We will use temp1 to store a subvector of prime. For each loop we initialize temp1 first in order to set it as an empty vector.
		vector<int> temp;
		//Generate prime sequence and its power first.
		vector<int> prime;
		vector<int> power_of_prime;
		
		//Set t_0 = 1.
		prime.push_back(1);
		power_of_prime.push_back(1);
		
		
		//Here we use the PrimeSequence function to generate prime and power_of_prime, which are vectors with the same length.
		PrimeSequence(prime, power_of_prime, 2);
		//Use row_sampling_scheme to generate a row of the sample scheme matrix and push it back into sample_scheme.
		row_sampling_scheme(sample_scheme, prime, power_of_prime);
	}
}



//Generate fftw_plans as a dictionary: map <int, fftw_plan> fftw_plans.
//Each element of fftw_plans is a pair, consisting of
//int: corresponds to the length of the DFT, and
//fftw_plan: stores the pointer of fftw_plan with the corresponding length.
void create_plans(map <int, fftw_plan>& fftw_plans, map <int, fftw_plan>& fftw_plans_back, vector<int>& length){
	
	if (fftwMeasure){
		fftw_complex *in, *out;
	    for (int i = 0; i < length.size(); ++i){
		    in = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    fftw_plans.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_FORWARD, FFTW_MEASURE)));
            fftw_plans_back.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_BACKWARD, FFTW_MEASURE)));
	    }
	}
	else{
		fftw_complex *in, *out;
	    for (int i = 0; i < length.size(); ++i){
		    in = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		    fftw_plans.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_FORWARD, FFTW_ESTIMATE)));
            fftw_plans_back.insert(pair<int, fftw_plan>(length[i], fftw_plan_dft_1d(length[i], in, out, FFTW_BACKWARD, FFTW_ESTIMATE)));
	    }
	}
}

//Fourier space to time space.
//A_hat ---> samples
void create_sample_time(vector<fftw_complex*>& samples, vector<int>& length, map <int, fftw_complex*>& A_hat, map <int, fftw_plan>& fftw_plans_back){
	complex<double> noise;
	double u1, v1;
	
	for (int i = 0; i < length.size(); ++i){
		fftw_execute_dft(fftw_plans_back[length[i]], A_hat[length[i]], samples[i]);
		/*
		//Add noise.
		for (int j = 0; j < length[i]; ++j){
			u1 = drand48();
            v1 = drand48();
	        //Create random noise.
	        noise = std_noise * sqrt(-2*log(u1)) * exp(2*M_PI * I * v1);
			
			samples[i][j][0] += real(noise);
			samples[i][j][1] += imag(noise);
		}
		*/
	}
}


void add_noise(vector<fftw_complex*>& samples, vector<int>& length){
	//Assume the total number of elements in samples is m.
	int total_length = 0;
	double sig_power = 0;
	double noise_power = 0;
	//Evaluate the power of all the elements in samples.
	for(int i = 0; i < samples.size(); ++i){
		total_length  += length[i];
		for(int j = 0; j < length[i]; ++j){
			sig_power += (samples[i][j][0]*samples[i][j][0] + samples[i][j][1]*samples[i][j][1]);
		}
	}
	
	//Create a noise vector which contains m elements.
	vector<complex<double>> noise_vector;
	noise_vector.resize(total_length);
	//Each element is an i.i.d random Gaussian noise with mean 0 and variance 1.
	//They are created using the box muller method.
	double u1, v1;
	for(int i = 0; i < total_length; ++i){
		u1 = drand48();
        v1 = drand48();
		noise_vector[i] = sqrt(-2*log(u1)) * exp(2*M_PI * I * v1);
		//Calculate the power of noise vector.
		noise_power += -2*log(u1);
	}
	
	//Normalize the noise vector based on the desired SNR value, the power of noise vector
	//and the power of samples.
	int temp = 0;
	double norm_factor = sqrt(sig_power/double(pow(10, SNR/10)))/sqrt(noise_power);
	for(int i = 0; i < total_length; ++i){
		noise_vector[i] *= norm_factor;
	}
	
	/*
	//exam SNR
	double new_noise_power = 0;
	for(int i = 0; i < total_length; ++i){
		new_noise_power += (abs(noise_vector[i])*abs(noise_vector[i]));
	}
	//cout << "noise_power is: " << noise_power << endl;
	//cout << "norm_factor is: " << norm_factor << endl;
	//cout << "sig_power is: " << sig_power << endl;
	//cout << "new_noise_power is: " << new_noise_power << endl;
	cout << "The actual SNR is: " << 10*log10(sig_power/new_noise_power) << endl;;
	*/
	
	//Add the noise vector elementwise to samples.
	int noise_vector_index = 0;
	for(int i = 0; i < samples.size(); ++i){
		for(int j = 0; j < length[i]; ++j){
			samples[i][j][0] += real(noise_vector[noise_vector_index]);
			samples[i][j][1] += imag(noise_vector[noise_vector_index]);
			++noise_vector_index;
		}
	}
	
}




//Time space to Fourier space.
//samples ---> A_hat
void create_sample_fourier(vector<fftw_complex*>& samples, vector<int> length, map <int, fftw_complex*>& A_hat, map <int, fftw_plan>& fftw_plans){
	fftw_complex* temp;
	for (int i = 0; i < length.size(); ++i){
		temp = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		fftw_execute_dft(fftw_plans[length[i]], samples[i], temp);
		for(int j = 0; j < length[i]; ++j){
			temp[j][0] /= length[i];
			temp[j][1] /= length[i];
		}
		
		A_hat.insert(make_pair(length[i], temp));
	}
}


//Create a short sequence in Fourier space based on the given frequencies and coefficients.
void create_A_hat_stu(map <int, fftw_complex*>& DFT_samples, vector<int> length, vector<double>& frequency, vector<complex<double>>& coefficient){
	fftw_complex* temp;
	for(int i = 0; i < length.size(); ++i){
		temp = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * length[i]);
		for (int k = 0; k < length[i]; ++k){
			temp[k][0] = 0;
			temp[k][1] = 0;
		}
		//Fill temp.
		for (int j = 0; j < frequency.size(); ++j){
			if (int(frequency[j]) % length[i] >= 0){
				temp[int(frequency[j]) % length[i]][0] += real(coefficient[j]);
			    temp[int(frequency[j]) % length[i]][1] += imag(coefficient[j]);
			}
			else{
				temp[(int(frequency[j]) % length[i]) + length[i]][0] += real(coefficient[j]);
			    temp[(int(frequency[j]) % length[i]) + length[i]][1] += imag(coefficient[j]);
			}
		}
		DFT_samples.insert(make_pair(length[i], temp));
	}
}



//Calculate the Discrete Fourier Transforms using FFTW.
//input: vector<vector <complex<double>>>& samples (the vector containing all the samples)
//input: vector<vector <fftw_plan>> fftw_plans (the vector containing the fftw_plans)
//output: vector<vector <complex<double>>>& DFT_samples (the vector containing all the samples after Fourier Tansform)
void DFT(vector<fftw_complex*>& samples, map <int, fftw_plan>& fftw_plans, map <int, fftw_complex*>& DFT_samples, vector<int>& sample_scheme_flat) {
	fftw_complex *in, *out;
	//For each DFT
	//1. define in and out
	//2. record the length of each vector that will take DFT
	//3. execute the corresponding fftw_plan to get out
	//4. store out in DFT_samples
	int dft_length;
	for (int i = 0; i < samples.size(); ++i) {
		dft_length = sample_scheme_flat[i];
		//1. define in and out
		in = samples[i];
		out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * dft_length);
		
		//2. fill in with complex numbers
		for (int l = 0; l < dft_length; ++l){
			in[l][0] = in[l][0]/dft_length;
			in[l][1] = in[l][1]/dft_length;
		}
		fftw_execute_dft(fftw_plans[dft_length], in, out);
		DFT_samples.insert(make_pair(dft_length, out));
	}
}


//Define a new function that compares two pairs of the form <double number, index(integer)>.
bool mag_complex(pair <complex<double>, int> a, pair <complex<double>, int> b)
{
	return abs(a.first) > abs(b.first);
}


//Extended Euclidean algorithm.
long long int extended_gcd(int a, int b){
	long long int s = 0;
	long long int old_s = 1;
	long long int t = 1;
	long long int old_t = 0;
	long long int r = b;
	long long int old_r = a;
	long long int quotient;
	long long int prov;
	while(r != 0){
		quotient = old_r / r;
		
		prov = r;
		r = old_r - quotient * prov;
		old_r = prov;
		
		prov = s;
		s = old_s - quotient * prov;
		old_s = prov;
		
		prov = t;
		t = old_t - quotient * prov;
		old_t = prov;
	}
    
	return old_s;
}

//Reconstruct the frequencies from their residues r modulo pairwise relatively prime numbers p.
int reconstruction(vector <int>& r, vector <int>& p){
	int m = r.size() - 1;
	long long int nu = r[0];
	long long int n = p[0];
	int W;
	
	for (int l = 1; l < m + 1; ++l){
		W = extended_gcd(n, p[l]);
		
		if ( ((r[l]-nu)*W) % p[l] >= 0){
			nu += n*(((r[l]-nu)*W) % p[l]);
		}
		else{
			nu += n*(((r[l]-nu)*W) % p[l] + p[l]);
		}
		
		n = p[l] * n;
	}
	
	if (0 <= nu % n and nu % n <= floor(n/2)){
		nu = nu % n;
	}
	else{
		nu = nu % n - n;
	}
	while (nu > floor(N/2)){
		nu -= N;
	}
	while (nu < -(floor(N/2) + 1) + 1){
		nu += N;
	}
	
	return nu;
}




//Find the median of the entries of a given vector.
double findMedian(vector<double> vec) {
	double median;
	int size_vec = vec.size();

	sort(vec.begin(), vec.end());

	if (size_vec % 2 == 0) {
		median = (vec[size_vec / 2 - 1] + vec[size_vec / 2]) / 2.0;
	}
	else {
		median = vec[size_vec / 2];
	}
	return median;
}


//Find element of s_k_t_l_u which has the closest magnitude compared to s_k_u[res].
int find_b_min(fftw_complex* s_k_u, fftw_complex* s_k_t_l_u, int tl, int res, int s_ku){
	double temp, temp1;
	int index_b = 0;
	temp = (s_k_u[res][0] - s_k_t_l_u[res][0])*(s_k_u[res][0] - s_k_t_l_u[res][0]) + (s_k_u[res][1] - s_k_t_l_u[res][1])*(s_k_u[res][1] - s_k_t_l_u[res][1]);
	
	for (int b = 1; b < tl; ++b){
		temp1 = (s_k_u[res][0] - s_k_t_l_u[res+b*s_ku][0])*(s_k_u[res][0] - s_k_t_l_u[res+b*s_ku][0]) + (s_k_u[res][1] - s_k_t_l_u[res+b*s_ku][1])*(s_k_u[res][1] - s_k_t_l_u[res+b*s_ku][1]);
		if (temp1 < temp){
			temp = temp1;
			index_b = b;
		}
	}
	return index_b;
}


//Define a new function that compares two pairs in R.
bool R_mag_complex(pair<int, complex<double>> a, pair<int, complex<double>> b)
{
	return abs(a.second) > abs(b.second);
}

//Define a new function that compares two pairs in R.
bool R_freq(pair<int, complex<double>> a, pair<int, complex<double>> b)
{
	return a.first > b.first;
}


//SSFT algorithm for block-structured sparse functions.
void multi_block_alg(vector<int>& w, vector<int>& t_l, vector <pair<int, complex<double>>>& R, map <int, fftw_complex*>& DFT_samples){
	//Each element of A_sort is a pair where the first entry is a complex number, and the second entry is the index based on sort result.
	int tl = t_l.size();
	//r will be used to store the residue.
	vector <int> r;
	r.resize(tl);
	//num stores the moduli for the CRT reconstruction procedure.
	vector <int> num;
	num.resize(tl);
	//Initialize num.
	for(int i = 1; i < tl; ++i){
		num[i] = t_l[i];
	}
	
	int temp_C;
	
	map <int, int> C;
	
	vector <pair<int, complex<double>>> R_temp;
	
	vector <double> real_part;
	vector <double> imag_part;
	real_part.resize(s_k.size());
	imag_part.resize(s_k.size());
	
	complex<double> temp_coeff;
	
	int b_min;
	
	int count_cout = 0;
	
	for (int v = 0; v < u; ++v){
		for (int k = 0; k < s_k.size(); ++k){
			num[0] = s_k[k] * u;
		    //We already know the value of w (it was pre-computed).
			for (int h = 0; h < s_k[k]; ++h){
				
				if ((h-v)*w[k] % s_k[k] >= 0){
					r[0] = (((h-v)*w[k] % s_k[k]) * u + v);
				}
				else{
					r[0] = ((((h-v)*w[k] % s_k[k]) + s_k[k]) * u + v);
				}
				
				for (int l = 1; l < tl; ++l){
					b_min = find_b_min(DFT_samples[s_k[k] * u], DFT_samples[s_k[k] * u * t_l[l]], t_l[l], r[0], s_k[k]*u);
					
					r[l] = (r[0] + b_min * s_k[k] * u) % t_l[l];
				}
				//Reconstruct the frequencies.
				temp_C = reconstruction(r, num);
				
			    if (C.find(temp_C) == C.end()) {
				    //If we did not find this frequency yet, add it.
				    C.insert(pair<int, int>(temp_C, 1));
			    }
			    else {
				    //If we found the frequency before:
				    C[temp_C] += 1;
			    }
			}
		}
		
		//Fourier Coefficient Estimation
		for (auto const& item : C){
			//Only consider elements which were reconstructed more than K/2 times.
			if (item.second > K/2){
				for(int i = 0; i < s_k.size(); ++i){
					if ((item.first % (s_k[i]*u*t_l[tl-1])) >= 0){
						real_part[i] = DFT_samples[s_k[i]*u*t_l[tl-1]][item.first % (s_k[i]*u*t_l[tl-1])][0];
					    imag_part[i] = DFT_samples[s_k[i]*u*t_l[tl-1]][item.first % (s_k[i]*u*t_l[tl-1])][1];
					}
					else{
						real_part[i] = DFT_samples[s_k[i]*u*t_l[tl-1]][(item.first % (s_k[i]*u*t_l[tl-1])) + s_k[i]*u*t_l[tl-1]][0];
					    imag_part[i] = DFT_samples[s_k[i]*u*t_l[tl-1]][(item.first % (s_k[i]*u*t_l[tl-1])) + s_k[i]*u*t_l[tl-1]][1];
					}
					
				}
				temp_coeff = (findMedian(real_part) + findMedian(imag_part)*I);
				
				R_temp.push_back(pair<int, complex<double>>(item.first, temp_coeff));
			}
		}
		
		sort(R_temp.begin(), R_temp.end(), R_mag_complex);
		//Select at most 2*num_block elements with the largest magnitude from R_temp and store them in R.
		if(R_temp.size() > 2*num_block){
			for (int i = 0; i < 2*num_block; ++i){
			    R.push_back(R_temp[i]);
		    }
		}
		else{
			for (int i = 0; i < R_temp.size(); ++i){
			    R.push_back(R_temp[i]);
		    }
		}
		C.clear();
		R_temp.clear();
	}
}



//Check the correctness of the recovery, i.e., whether the algorithm found all significant elements or not.
bool check_correct_recovery(vector <pair<int, complex<double>>>& R_est, vector <pair<int, complex<double>>>& R_true){
	for (int i = 0; i < num_block*B; ++i){
		if (R_est[i].first != R_true[i].first){
			return false;
		}
	}
	return true;
}


//For correct recovery, calculate the absolute value of the error for each frequency.
void err_estimate(vector <pair<int, complex<double>>>& R_est, vector <pair<int, complex<double>>>& R_true, vector <double>& err){
	for (int i = 0; i < num_block*B; ++i){
		err.push_back(abs(R_est[i].second - R_true[i].second));
	}
}

//Calculate the error in average L1 norm.
double average_l1_err(vector <double> err){
	double temp_sum = 0;
	for (int i = 0; i < err.size(); ++i){
		temp_sum += err[i];
	}
	return temp_sum / err.size();
}




int main()
{
	if(!fftwMeasure){
	    cout << "Using FFTW_ESTIMATE, not FFTW_MEASURE!" << endl;
    }
	cout << "The bandwidth, number of blocks and frequencies in each block are: " << N << ", " << num_block << ", " << B << endl;
	
	//Find the smallest u.
	u = smallest_u(2);
	
	cout << "u: " << u << endl;
	cout << "The number of blocks is: " << num_block << endl;
	
	//Find t_l.
	
	//Determine the value of t_l.
	vector<vector <int> > t_l;
	
	//t_l contains power of prime.
	t_l_info(t_l);
	
	/*
	//t_l contains prime numbers only.
	vector<int> temp_t_l;
	temp_t_l.push_back(1);
	int temp_product = 1;
	int temp_index_t_l = 0;
	while (temp_product < N / double(u) / num_block){
		++temp_index_t_l;
		temp_t_l.push_back(PRIMES[temp_index_t_l]);
		temp_product *= PRIMES[temp_index_t_l];
	}
	t_l.push_back(temp_t_l);
	*/
	
	
	cout << "t_l" << endl;
	int t_L = 0;
	for (int i = 0; i < t_l.size(); ++i){
		for (int j = 0; j < t_l[i].size(); ++j){
			if (t_L < t_l[i][j]){
				t_L = t_l[i][j];
			}
			cout << t_l[i][j] << " ";
		}
		cout << endl;
	}
	
	//Find the first elements of s_k.
	int temp_index = 0;
	int temp = PRIMES[temp_index];
	while (temp <= num_block or temp <= t_L){
		++temp_index;
		temp = PRIMES[temp_index];
	}
	
	s_k.push_back(temp);
	
	//If the randomized version of the algorithm is being used, set K = 1.5 * log(2*num_block) + randomize_int.
	if (Randomized){
		K = 1.5 * log(2*num_block) + randomize_int;
	}
	else{
		K = 2*num_block*floor(log(N / double(u)) / log(s_k[0])) + 1;
	}
	
	
	for (int i = 1; i < K; ++i){
		++temp_index;
		s_k.push_back(PRIMES[temp_index]);
	}
	cout << "s_k" << endl;
	for (int i = 0; i < s_k.size(); ++i){
	    cout << s_k[i] << " ";
	}
	cout << endl;
	
	
	//Calculate the total number of samples that will be used in the algorithm.
	int total_sample = 0;
	for (int k = 0; k < s_k.size(); ++k){
		for (int l = 0; l < t_l[0].size(); ++l){
			total_sample += s_k[k] * t_l[0][l] * u;
		}
	}
	cout << "The ratio between the total number of samples and the bandwidth: " << total_sample / double(N) << endl;
	
	
	//Compute the extended_gcd for each pair of s_k and u and store the results in w.
	vector<int> w;
	for (int k = 0; k < s_k.size(); ++k){
		w.push_back(extended_gcd(u, s_k[k]));
	}
	
	vector<int> sample_length;
	for (int k = 0; k < s_k.size(); ++k){
	    for (int l = 0; l < t_l[0].size(); ++l){
		    sample_length.push_back(s_k[k] * t_l[0][l] * u);
	    }
	}
	
	
	cout << "Initialize samples..." << endl;
	//Initialize samples.
	vector<fftw_complex*> samples;
	fftw_complex *samples_temp;
	for (int k = 0; k < sample_length.size(); ++k){
		samples_temp = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * sample_length[k]);
		samples.push_back(samples_temp);
	}
	
	
	clock_t start;
	
	start = std::clock();
	
	
	cout << "Create fftw plans for forward and backward FFT" << endl;
	//Compute the fftw_plan for different lengths of samples.
	map <int, fftw_plan> fftw_plans;
	map <int, fftw_plan> fftw_plans_backward;
	//int stores the length of each fftw_plan. Since all occurring numbers are pairwise relatively prime, each length must be unique.
	create_plans(fftw_plans, fftw_plans_backward, sample_length);	
	
	cout << "Create fftw_plans uses: (in second) " << ( clock() - start ) / (double) CLOCKS_PER_SEC << endl;

	//Now all precomputations are completed.
	
	cout << "Start numerical experiment..." << endl;
	
	double duration;
	
	int seed = 123;
	
	vector <double> err;
	int correct = 0;
	//Different numerical tests.
	cout << "Total number of different tests is: " << rounds << ". " << endl;
	cout << "Each test is repeated " << num_test << " times." << endl;
	for (int round = 0; round < rounds; ++round){
		
		//cout << round << endl;
		
		srand(seed);
	    ++seed;
		
		//Create frequencies and coefficients.
		//The easiest setting is to pick num_block numbers from [0, N).
		vector<double> freq_start;
		vector<double> freq;
		vector<complex <double> > coeff;
		double arg;
		
		bool bad_freqs = true;
		
		//We do not want the elements of freq_start to be too close to each other.
		while(bad_freqs){
			freq_start.clear();
			for (int i = 0; i < num_block; i++)
			    freq_start.push_back(rand() % (N - B - 1) - N/2 + 1);
		    sort(freq_start.begin(), freq_start.end());
			if(num_block == 1){
				break;
			}
			for(int j = 0; j < num_block - 1; ++j){
				if(freq_start[j + 1] - freq_start[j] < B + 1){
					bad_freqs = true;
					break;
				}
				if(j == num_block - 2 and freq_start[j + 1] - freq_start[j] >= B + 1){
					bad_freqs = false;
				}
			}
		}
		
		
		//Create frequencies and their corresponding coefficients.
		for (int i = 0; i < num_block; ++i){
			for (int bb = 0; bb < B; ++bb){
				freq.push_back(freq_start[i] + bb);
				arg = 2 * pi * drand48();
				coeff.push_back(complex<double>(cos(arg), sin(arg)));
			}
		}
		
		//Create A_hat, then use A_hat to create the required samples.
		map <int, fftw_complex*> A_hat;
		
		//cout << "Calculate A_hat..." << endl;
		create_A_hat_stu(A_hat, sample_length, freq, coeff);
		
		
		//cout << "Create required samples in time space..." << endl;
		//Create the required samples.
		create_sample_time(samples, sample_length, A_hat, fftw_plans_backward);
		
		if(!noiseless){
			add_noise(samples, sample_length);
		}
		
		vector <pair<int, complex<double>>> R;
		
		clock_t start;
	
	    start = std::clock();
		
		//Repeat the numerical tests (for timing purposes).
		for(int test = 0; test < num_test; ++test){
			
			R.clear();
			
			//Apply DSFT to each samples vector and store them in an array that has the same data structure as A_hat.
			map <int, fftw_complex*> samples_hat;
			
			//cout << "Create required samples in Fourier space..." << endl;
			create_sample_fourier(samples, sample_length, samples_hat, fftw_plans);
						
			multi_block_alg(w, t_l[0], R, samples_hat);
			
		}
		
		duration = ( clock() - start ) / (double) CLOCKS_PER_SEC / num_test;
		
		//Check the correctness and calculate the error for each recovery.
		
		vector <pair<int, complex<double>>> R_est;
		vector <pair<int, complex<double>>> R_true;
		
		//Sort R based on the magnitude of its elements.
	    sort(R.begin(), R.end(), R_mag_complex);
		
		
	    for (int i = 0; i < num_block*B; ++i){
		    R_est.push_back(R[i]);
	    }
		
		for (int i = 0; i < num_block*B; ++i){
			R_true.push_back(pair<int, complex<double>>(freq[i], coeff[i]));
		}
		
		sort(R_est.begin(), R_est.end(), R_freq);
	    sort(R_true.begin(), R_true.end(), R_freq);
		
		//Compare R_true and R_est to check the correctness.
		bool correct_recovery = false;
		correct_recovery = check_correct_recovery(R_est, R_true);
		
		if (correct_recovery){
			correct += 1;
			err_estimate(R_est, R_true, err);
		}
	}
	
	//Error analysis
	//Maximum error and average L1 norm.
	sort(err.begin(), err.end(), greater<double>());
	
	double average_l1 = average_l1_err(err);
	
	//cout << "The bandwidth, number of blocks and frequencies in each block are: " << N << ", " << num_block << ", " << B << endl;
	if (!noiseless){
	    cout << "The noise level (SNR) is: " << SNR << endl;
	}
	else{
		cout << "There is no noise in the original signal." << endl;
	}
	cout << "Among " << rounds << " numerical tests the number of correct recovery is " << correct << endl;
	cout << "The maximum error among the numerical tests is : " << err[0] << endl;
	cout << "The error in average L1 norm is : " << average_l1 << endl;
	cout << "The running time is: " << duration << endl;
	
	
	return 0;
}
